/*
NDS Translucent Plasma
Relminator

http://rel.betterwebber.com

*/

#include <nds.h>
#include <stdio.h>
#include <math.h>
#include <string.h>




inline void cls(u16 color)
{
	//dmaFillWords (u32 value, void *dest, uint32 size)
	dmaFillWords (color, VRAM_A, SCREEN_HEIGHT * SCREEN_WIDTH * 2);
}

inline void pcopy(u16 *buffer1,u16 *buffer2)
{
	//dmaCopy (const void *source, void *dest, uint32 size)
	//dmaCopy (buffer1, buffer2, SCREEN_HEIGHT * SCREEN_WIDTH*2);
	
	memcpy(buffer2, buffer1, SCREEN_HEIGHT * SCREEN_WIDTH*sizeof(u16));
}

inline void pset(s32 x, s32 y, u16 color)
{
	if (x<0) return;
	if (y<0) return;
	if (x>=SCREEN_WIDTH) return;
	if (y>=SCREEN_HEIGHT) return;
	
	VRAM_A[(y * SCREEN_WIDTH) + x] = color;
}    

void lens(u16 *page1,u16 *page2,s32 x, s32 y, s32 radius)
{

   	const int SCR_X_MAX = SCREEN_WIDTH - 1;
	const int SCR_Y_MAX = SCREEN_HEIGHT - 1;

    s32 wid, hei;
    s32 hypotsquared, radiussquared, h;
    s32 sx, sy, x1, y1, yt, xt;
    s32 px, py;
    s32 minx, miny;
    s32 wtemp, htemp;
    s32 sphereheight, cleaner;
    


    sphereheight = (radius >> 1);
    cleaner = sphereheight * 10;

    radiussquared = radius * radius;

    wid = radius << 1;
    hei = wid;


    minx = 0;
    miny = 0;

  	if (y < 0)
	{
		y = -y;
		miny = y;
		hei = hei - y;
		if (hei <= 0) return;
		y = 0;
	}

	if  ((y + hei) > SCR_Y_MAX)
	{
		htemp = (y + hei) - SCREEN_HEIGHT;
		hei = hei - htemp;
		if (hei <= 0) return;
	}

	if (x < 0)
	{
		x = -x;
		minx = x;
		wid = wid - x;
		if (wid <= 0) return;
		x = 0;
	}
	
	if  ((x + wid) > SCR_X_MAX)
	{
		wtemp = (x + wid) - SCREEN_WIDTH;
		wid = wid - wtemp;
		if (wid <= 0) return;
	}
        
    
    for (yt=0; yt<hei; yt++)
	{
		for (xt=0; xt<wid; xt++)
		{
			x1 = (xt - radius) + minx;
			y1 = (yt - radius) + miny;

			hypotsquared = (x1 * x1) + (y1 * y1);
			if (hypotsquared < (radiussquared - cleaner))
			{
				h = (sqrt32(radiussquared - hypotsquared));
				sx = (sphereheight - h);
				sy = (sphereheight - h);
				sx = sx * (div32(x1 << 16, h));  //div32 here
				sy = sy * (div32(y1 << 16, h));  // ditto 
				py = ((sy >> 16) + yt + y);
				px = ((sx >> 16) + xt + x);
				sx = (x+xt);
				sy = (y+yt);
				if ( (px > -1) && (px < SCREEN_WIDTH) && (py > -1) && (py < SCREEN_HEIGHT) )                  
				{
					page2[(sy * SCREEN_WIDTH) + sx] = page1[(py * SCREEN_WIDTH) + px];              
				}                
			}                                    
		}       
    }

}


//*************************************************************************************
//
//   
//
//*************************************************************************************
// some sine LUTs
u8 lsin1[2048];
u8 lsin2[2048];
u8 lsin3[2048];

// RGB palette
u16 pal[256]; 

// lens buffer and double buffer
u16 vpage1[SCREEN_HEIGHT*SCREEN_WIDTH];
u16 vpage2[SCREEN_HEIGHT*SCREEN_WIDTH];
   

int main(void)
{
	
	// constant sine divisor
	const u8 K1 = 32;
    const u8 K2 = 12;
    const u8 K3 = 28;
	
	// maximum values our dynamic plasmas can move
	u16 KSIN1 = u8(K1 *(360/57.3f));
    u16 KSIN2 = u8(K2 *(360/57.3f));
    u16 KSIN3 = u8(K3 *(360/57.3f));
	
	
    		
	// precalculate our luts
	for (int i = 0; i< ((2048) - 1); i++)
	{
        lsin1[i] = sin(i/(float)K1) * 32+32;
        lsin2[i] = sin(i/(float)K2) * 16+16;
        lsin3[i] = sin(i/(float)K3) * 20+20;
    }    

	// generate our 256 color palette
	for (int i=0; i<256; i++)
	{
	    u8 r = (u8)(abs(int(16 - 15 * sin(i * M_PI / 12.0f))));
		u8 g = (u8)(abs(int(16 - 15 * sin(i * M_PI / 16.0f))));
		u8 b = (u8)(abs(int(16 - 15 * sin(i * M_PI / 8.0f))));
		pal[i] = RGB15(r,g,b);
	}

	irqInit();
	irqEnable(IRQ_VBLANK);
	
	
	//initialize the DS Dos-like functionality
	consoleDemoInit();
	
	iprintf("\x1b[1;1HLensmapping + Plasma!");
	iprintf("\x1b[2;1HFramebuffer madness!");
	iprintf("\x1b[3;1HRelsoft");
	iprintf("\x1b[4;1Hhttp://rel.betterwebber.com");
	iprintf("\x1b[6;1HAnya Therese B. Lope");
 
	//set frame buffer mode 0
	videoSetMode(MODE_FB0);
	
	
 
	//enable VRAM A for writting by the cpu and use 
	//as a framebuffer by video hardware
	vramSetBankA(VRAM_A_LCD);
 
	
	u16 a = 0, b = 0, ct = 0;	// movement deltas
	u16 c;						// palette color index
	
	
	// lens params
	s32 lens_x=90;
	s32 lens_y=80;
	s32 lens_rad=40;
	s32 angle = 0;
	
	
	
	while(1)
	{

		// move our lens in a funky way
		angle += 180;
		
		
		lens_y = (cosLerp(angle*6)+sinLerp(angle))*30;
		lens_x = (sinLerp(angle+1024)+sinLerp(-angle*2))*30;
		
		lens_x = (lens_x >> 12) + SCREEN_WIDTH/2;
		lens_y = (lens_y >> 12) + SCREEN_HEIGHT/2; 
		
		
		
		// dynamically move our plasmas 
		a= (a + 1) % (KSIN1 + 1);
		b = (b + 1) % (KSIN2 + 1);
		ct = (ct + 1) % (KSIN3 + 1);
		
		
		
		// Generate plasma
		// write to page 1 directly
		u16 *p_offset = vpage1;
		
		for (int ya = 0; ya<SCREEN_HEIGHT; ya++)
		{
			for (int xa = 0; xa <SCREEN_WIDTH; xa++)
			{
				
				// calculate colors
				c = (lsin1[xa + a ] + lsin2[ya + b ] +lsin3[1024+xa + ya - ct]);
				c = (lsin1[xa + a +c ] + lsin2[ya + b + c ] +lsin3[1024+xa + ya - ct+c]);
				// write to vram
				*p_offset++ = pal[c];
			}
			
		}		
		
		// copy palsma to lensbuffer
		pcopy(vpage1,vpage2);
		
		// lens it
		lens(vpage1, vpage2,lens_x-lens_rad,lens_y-lens_rad,lens_rad);
		
		// copy the lensed buffer to vram
		pcopy(vpage2,VRAM_A);
		
		
		swiWaitForVBlank();
	}
 
	return 0;
}
 